# -*- coding: utf-8 -*-
## A Plone product to generate quiz. A quiz is a list of questions with the appropriate answer(s).
## Content managers can build quiz easily and present users a report about the errors they did.
## Copyright (C)2005 Ingeniweb

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; see the file COPYING. If not, write to the
## Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
"""
A Plone product to generate quiz. A quiz is a list of questions with the appropriate answer(s)
"""
__docformat__ = 'restructuredtext'


# Python imports
from math import floor

# Zope imports
from AccessControl import ClassSecurityInfo

# CMF imports
from Products.CMFCore import CMFCorePermissions
from Products.CMFCore.utils import getToolByName

# Archetypes imports
try:
  from Products.LinguaPlone.public import *
except ImportError:
  # No multilingual support
  from Products.Archetypes.public import *
from Products.Archetypes.utils import make_uuid


# Other imports
try:
    from Products.ATContentTypes.types.ATFolder import ATFolder
except:
    # We use Plone 2.1
    from Products.ATContentTypes.content import ATFolder
from Products.PloneQuiz.content.quizqa import QuizQA
from Products.PloneQuiz.config import I18N_DOMAIN
from Products.PloneQuiz.content.vocabulary import SCORE_CATEGORY_LIST


# Attribute where all user scores are stored
SCORE_HISTORY_ATTRIBUTE = 'score_history'

QuizSchema = ATFolder.schema.copy() + Schema((
    BooleanField('showOnOnePage',
        default=False,
        accessor='isShownOnOnePage',
        widget=BooleanWidget(
            label='Show quiz on one page',
            label_msgid='label_show_quiz_on_one_page',
            description_msgid='help_show_quize_on_one_page',
            i18n_domain=I18N_DOMAIN,
        )
    ),
    BooleanField('immediateCorrection',
        default=True,
        accessor='useImmediateCorrection',
        widget = BooleanWidget(
            label='Immediate correction',
            label_msgid='label_immediate_correction',
            description_msgid='help_immediate_correction',
            i18n_domain=I18N_DOMAIN,
        )
    ),
    TextField('defaultRightAnswerMessage',
        schemata="configuration",
        widget=TextAreaWidget(
            label='Default right answer message',
            label_msgid='label_default_right_answer_message',
            description_msgid='help_default_right_answer_message',
            i18n_domain=I18N_DOMAIN,
        )
    ),
    TextField('defaultWrongAnswerMessage',
        schemata="configuration",
        widget=TextAreaWidget(
            label='Default wrong answer message',
            label_msgid='label_default_wrong_answer_message',
            description_msgid='help_default_wrong_answer_message',
            i18n_domain=I18N_DOMAIN,
        )
    ),
    StringField('veryGoodScoreTitle',
        schemata="scores",
        widget=StringWidget(
            label='Very good score title',
            label_msgid='label_very_good_score_title',
            description_msgid='help_very_good_score_title',
            i18n_domain=I18N_DOMAIN,
       ),
   ),
   StringField('veryGoodScoreMessage',
        schemata="scores",
        widget=TextAreaWidget(
            label='Very good score message',
            label_msgid='label_very_good_score_message',
            description_msgid='help_very_good_score_message',
            i18n_domain=I18N_DOMAIN,
       ),
   ),
       StringField('goodScoreTitle',
        schemata="scores",
        widget=StringWidget(
            label='Good score title',
            label_msgid='label_good_score_title',
            description_msgid='help_good_score_title',
            i18n_domain=I18N_DOMAIN,
       ),
   ),
   StringField('goodScoreMessage',
        schemata="scores",
        widget=TextAreaWidget(
            label='Good score message',
            label_msgid='label_good_score_message',
            description_msgid='help_good_score_message',
            i18n_domain=I18N_DOMAIN,
       ),
   ),
   StringField('badScoreTitle',
        schemata="scores",
        widget=StringWidget(
            label='Bad score title',
            label_msgid='label_bad_score_title',
            description_msgid='help_bad_score_title',
            i18n_domain=I18N_DOMAIN,
       ),
   ),
   StringField('badScoreMessage',
        schemata="scores",
        widget=TextAreaWidget(
            label='Bad score message',
            label_msgid='label_bad_score_message',
            description_msgid='help_bad_score_message',
            i18n_domain=I18N_DOMAIN,
       ),
   ),
   StringField('veryBadScoreTitle',
        schemata="scores",
        widget=StringWidget(
            label='Very bad score title',
            label_msgid='label_very_bad_score_title',
            description_msgid='help_very_bad_score_title',
            i18n_domain=I18N_DOMAIN,
       ),
   ),
   StringField('veryBadScoreMessage',
        schemata="scores",
        widget=TextAreaWidget(
            label='Very bad score message',
            label_msgid='label_very_bad_score_message',
            description_msgid='help_very_bad_score_message',
            i18n_domain=I18N_DOMAIN,
       ),
   ),
),
)

class Quiz(ATFolder):
    security = ClassSecurityInfo()

    __implements__ = ATFolder.__implements__

    archetype_name              = 'Quiz'
    meta_type                   = 'Quiz'
    portal_type                 = 'Quiz'
    global_allow                = True
    filter_content_types        = True
    allowed_content_types       = ('QuizQA',)
    allow_discussion            = False
    content_icon                = 'quiz.gif'
    immediate_view              = 'quiz_view'
    default_view                = 'quiz_view'
    schema = QuizSchema

    actions = ({
        'id'          : 'view',
        'name'        : 'View',
        'action'      : 'string:${folder_url}/',
        'permissions' : (CMFCorePermissions.View,)
         },)

    security.declareProtected(CMFCorePermissions.View, 'getQuiz')
    def getQuiz(self):
        """Returns quiz object itself"""
        
        return self
    
    security.declareProtected(CMFCorePermissions.View, 'getQuiz')
    def buildUserQuizUID(self):
        """Create and returns an uid used to identify a user playing the quiz.
        This uid is used to keep history of user score"""
        
        return make_uuid()
    
    security.declareProtected(CMFCorePermissions.View, 'getQuizQAs')
    def getQuizQAs(self):
        """Returns question/answer objects"""
        
        objs = self.objectValues()
        result = []
        for obj in objs:
            if isinstance(obj, QuizQA):
                result.append(obj)
        return result

    security.declareProtected(CMFCorePermissions.View, 'getQuizQA')
    def getQuizQA(self, uid):
        """Returns a QuizQA object
        
        @param uid: UID of a QuizQA object
        """

        atool = getToolByName(self, 'archetype_tool')
        return atool.getObject(uid)

    security.declareProtected(CMFCorePermissions.View, 'isEmpty')
    def isEmpty(self):
        """Returns True if no question has been defined"""
        
        return (len(self.getQuizQAs()) > 0)
        

    security.declareProtected(CMFCorePermissions.View, 'getQuizQAPosition')
    def getQuizQAPosition(self, uid):
        """Returns QuizQA position in the quiz.
        Assume 0 is the first Question/Answers
        
        @param uid: uid of the QuizQA object
        """
        
        quiz_qas = self.getQuizQAs()
        index = 0
        pos = -1
        for quiz_qa in quiz_qas:
            if uid != quiz_qa.UID():
                index += 1
                continue
            pos = index
            break
        return pos

    security.declareProtected(CMFCorePermissions.View, 'getStatistics')
    def getStatistics(self):
        """Returns a dictionnary containing score by category.
        
        Each dictionnary value return a tuple(score, ratio)
        veryGood-> (score, ratio)
        good-> (score, ratio)
        bad-> (score, ratio)
        veryBad-> (score, ratio)
        """
 
        # Get all history
        score_history = getattr(self, SCORE_HISTORY_ATTRIBUTE, {})
     
        # History only keep score ratio (0.0 to 1.0)
        # Create a dictionnary with 4 keys corresponding to veryGood, good, bad
        # and veryBad values
        stats_scores = {}
        for score_ratio in score_history.values():
            category = self._getScoreCategory(score_ratio)
            category_score = stats_scores.get(category, 0)
            stats_scores[category] = category_score + 1
        
        # Append ratio to score
        total = sum([x for x in stats_scores.values()])
        stats = {}
        if total == 0:
            stats = dict([(k, (v, 0.0)) for k, v in stats_scores.items()])
        else:
            stats = dict([(k, (v, float(v)/total)) for k, v in stats_scores.items()])
            
        return stats
    
    security.declareProtected(CMFCorePermissions.View, 'getResult')
    def getResult(self):
        """Extract user results from request and returns a list of 
        dictionnaries. Each dictionnary contains user answers for a question.
        - uid : uid of QuizQA object
        - answer_ids: ids of user answers
        - valid: Are answers valid ?"""
        
        result = []
        request = self.REQUEST
        quiz_qa_uids = request.get('quiz_qa_uids', ())
        for uid in quiz_qa_uids:
            quiz_qa = self.getQuizQA(uid)
            result.append(quiz_qa.getResult())
        return result

    def _getScoreCategory(self, score_ratio):
        """Returns the category corresponding to the specified score ratio
        
        0.0<=score_ratio<0.25 equals to a very bad result
        0.25<=score_ratio<0.5 equals to a bad result
        0.5<=score_ratio<0.75 equals to a good result
        0.75<=score_ratio<1.0 equals to a very good result
        
        @param score_ratio: Score ratio (0.0 to 1.0) used to determine the 
        score category"""
        
        category = 'veryGood'
        if 0.0<=score_ratio<0.25:
            category = 'veryBad'
        if 0.25<=score_ratio<0.5:
            category = 'bad'
        if 0.5<=score_ratio<0.75:
            category = 'good'
        
        return category

    security.declareProtected(CMFCorePermissions.View, 'getScore')
    def getScore(self):
        """Returns a tuple :
        (correct answers, total answers, score ratio)"""
        
        result = self.getResult()
        correct_answers = len([x for x in result if x['valid']])
        total_answers = len(result)
        score_ratio = float(correct_answers)/total_answers
        score_category = self._getScoreCategory(score_ratio)
        return (correct_answers, total_answers, score_ratio, score_category)
    
    security.declareProtected(CMFCorePermissions.View, 'getScoreMessage')
    def getScoreMessage(self, category):
        """Returns the message corresponding to the specified category
        
        @param category: Category of score
        """
        
        field_name = '%sScoreMessage' % category
        field = self.getField(field_name)
        accessor = field.getAccessor(self)
        if accessor is None:
            return ""
        return accessor()
        
    security.declareProtected(CMFCorePermissions.View, 'getScoreTitle')
    def getScoreTitle(self, category):
        """Returns the title corresponding to the specified category
        
        @param category: Category of score
        """
        
        field_name = '%sScoreTitle' % category
        field = self.getField(field_name)
        accessor = field.getAccessor(self)
        if accessor is None:
            return ""
        title = accessor()
        
        # If quiz title is empty, use the vocabulary one
        if not title:
            vocab = self.getScoreCategoryVocabulary()
            title = vocab.getValue(category)
        
        return title
        

    security.declareProtected(CMFCorePermissions.View, 'storeResult')
    def storeResult(self):
        """Extract user results on quiz and store it in a quiz attribute
        """
        
        # Get key we will use to store user score
        request = self.REQUEST
        user_quiz_uid = request.get('user_quiz_uid')
        
        # Get score of user
        # We only store ratio
        score = self.getScore()
        score_ratio = score[2]
        
        # Get all history
        score_history = getattr(self, SCORE_HISTORY_ATTRIBUTE, {})
        
        # Stop udpating history if user score has already been stored
        if score_history.has_key(user_quiz_uid):
            return
        
        # Update score history
        score_history[user_quiz_uid] = score_ratio
        setattr(self, SCORE_HISTORY_ATTRIBUTE, score_history)
        
    security.declareProtected(CMFCorePermissions.View, 'isUserScoreStored')
    def isUserScoreStored(self, user_quiz_uid=None):
        """Returns true if user score has already been stored.
        
        @param user_quiz_uid: UID od user quiz. If not mentionned get it from 
        request
        """
        
        # Get user quiz uid from request if not defined
        if user_quiz_uid is None:
            if not request.has_key('user_quiz_uid'):
                return False
            user_quiz_uid = request.get('user_quiz_uid')
            
        # Get all history and check if user score has been stored
        score_history = getattr(self, SCORE_HISTORY_ATTRIBUTE, {})
        if score_history.has_key(user_quiz_uid):
            return True
        
        return False

    security.declareProtected(CMFCorePermissions.View, 'getAnswerMessage')
    def getAnswerMessage(self, is_valid):
        """Returns answer message. Use quiz one if not defined on QuizQA
        
        @param is_valid: Returns right message if this parameter is True 
            otherwise returns wrong message
        """
        
        if is_valid:
            # Returns the message for right answer
            msg = self.getDefaultRightAnswerMessage()
        else:
            # Returns the message for wrong answer
            msg = self.getDefaultWrongAnswerMessage()
        
        if not msg:
            return ''
        
        return msg
    
    security.declareProtected(CMFCorePermissions.View, 'getScoreCategoryVocabulary')
    def getScoreCategoryVocabulary(self):
        """Returns score category vocabulary"""
        
        return Vocabulary(SCORE_CATEGORY_LIST, self, I18N_DOMAIN)

#===============================================================================
#    GUI helpers used on quiz_view template
#===============================================================================

    security.declareProtected(CMFCorePermissions.View, 'getQuizQAsToDisplay')
    def getQuizQAsToDisplay(self):
        """Returns a list of QuizQAs to display on quiz_view template.
        This is a gui helper. It would be better to use a Z3 view"""
    
        quiz_qas = self.getQuizQAs()
        if not quiz_qas:
            return ()
        
        # Display one QuizQA object per page
        request = self.REQUEST
        quiz_qa_uids = request.get('quiz_qa_uids', ())
        quiz_qa_uids_length = len(quiz_qa_uids)
        
        # Check if we show correction
        if not self.isShownOnOnePage() and \
           request.get('show_correction', False) and \
           quiz_qa_uids_length > 0:
            return (quiz_qas[(quiz_qa_uids_length - 1)],)
        
        # User has answer all questions
        if quiz_qa_uids_length == len(quiz_qas):
            return ()
        
        # Display all QuizQA objects
        if self.isShownOnOnePage():
            return quiz_qas
        
        # Returns the next QuizQAs object
        return (quiz_qas[quiz_qa_uids_length],)
    
    security.declareProtected(CMFCorePermissions.View, 'displayResults')
    def displayResults(self):
        """Returns true if form must display the results button
        """
    
        quiz_qas = self.getQuizQAs()
        if not quiz_qas:
            return False
        
        if self.isShownOnOnePage():
            return True
        
        # Display one QuizQA object per page
        request = self.REQUEST
        quiz_qa_uids = request.get('quiz_qa_uids', ())
        if self.useImmediateCorrection() and \
           len(quiz_qas) == len(quiz_qa_uids):
            return True
        
        if not self.useImmediateCorrection() and \
           len(quiz_qas) == (len(quiz_qa_uids) + 1):
               return True
        
        return False
        
    security.declareProtected(CMFCorePermissions.View, 'buildUserAnswersAsHiddenInput')
    def buildUserAnswersAsHiddenInput(self):
        """Get all user answers and convert them to html hidden input tag"""
        
        inputs = []
        request = self.REQUEST
        html_input = '<input type="hidden" name="%(name)s" value="%(value)s" />'
        list_key = 'quiz_qa_uids'
        list_input_name = '%s:list' % list_key
        quiz_qa_uids = request.get(list_key, ())
        for uid in quiz_qa_uids:
            key = 'quiz_qa_%s' % uid
            answers = request.get(key, ())
            
            # Build quiz_qa_uids input
            html = html_input % {
                        'name': list_input_name,
                        'value': uid,}
            inputs.append(html)
            
            # Build quiz_qa_<uid> inputs
            input_name = '%s:int:list' % key
            for answer in answers:
                html = html_input % {
                            'name': input_name,
                            'value': answer,}
                inputs.append(html)
        
        return ' '.join(inputs)

registerType(Quiz)