# -*- coding: utf-8 -*-
## A Plone product to generate quiz. A quiz is a list of questions with the appropriate answer(s).
## Content managers can build quiz easily and present users a report about the errors they did.
## Copyright (C)2005 Ingeniweb

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; see the file COPYING. If not, write to the
## Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
"""
A Plone product to generate quiz. A quiz is a list of questions with the appropriate answer(s)
"""
__docformat__ = 'restructuredtext'

# Zope imports
from AccessControl import ClassSecurityInfo

# CMF imports
from Products.CMFCore import CMFCorePermissions

# Archetypes imports
from Products.Archetypes.public import *

# Other imports
from Products.ATReferenceBrowserWidget.ATReferenceBrowserWidget import \
    ReferenceBrowserWidget
from Products.PloneQuiz.config import CUSTOM_THUMB_SIZE, I18N_DOMAIN
from Products.PloneQuiz.field import AnswerField
from Products.PloneQuiz.widget import AnswerWidget

QuizQASchema = BaseSchema.copy() + Schema((
    AnswerField(
        'answers',
        required = True,
        widget=AnswerWidget(
            label='Answers',
            label_msgid='label_answers',
            description_msgid='help_answers',
            i18n_domain=I18N_DOMAIN,
        ),
    ),
    TextField(
        'rightAnswerMessage',
        widget=TextAreaWidget(
            label='Right answer message',
            label_msgid='label_right_answer_message',
            description_msgid='help_right_answer_message',
            i18n_domain=I18N_DOMAIN,

        ),
    ),
    TextField(
        'wrongAnswerMessage',
        widget=TextAreaWidget(
            label='Wrong answer message',
            label_msgid='label_wrong_answer_message',
            description_msgid='help_wrong_answer_message',
            i18n_domain=I18N_DOMAIN,
        ),
    ),

    StringField(
        'relatedUrl',
        validators=('isURL',),
        widget=StringWidget(
            label='Url',
            label_msgid='label_related_url',
            description_msgid='help_related_url',
            i18n_domain=I18N_DOMAIN,
        ),
    ),
    ReferenceField(
        'relatedContent',
        multiValued=False,
        allowed_types=(),
        relationship='relatesTo',
        vocabulary_display_path_bound=100,
        widget=ReferenceBrowserWidget(
            label='Related content',
            label_msgid='label_related_content',
            description_msgid='help_related_content',
            i18n_domain=I18N_DOMAIN,
            allow_search=False,
            force_close_on_insert=True,
        ),
    ),
    ImageField(
        'image',
        sizes={
            'mini': (80,80),
            'thumb': (150,150),
            'normal': (200,200),
            'big': (300,300),
            'custom': CUSTOM_THUMB_SIZE,},
        widget=ImageWidget(
            label='Image',
            label_msgid='label_image',
            description_msgid='help_image',
            i18n_domain=I18N_DOMAIN,
        ),
    ),

),
)

QuizQASchema['title'].widget.label = 'Question'
QuizQASchema['title'].widget.label_msgid = 'label_question'
QuizQASchema['title'].widget.i18n_domain = I18N_DOMAIN

class QuizQA(BaseContent):
    security = ClassSecurityInfo()
    __implements__ = BaseContent.__implements__

    archetype_name             = 'Question/Answer(s)'
    meta_type                  = 'QuizQA'
    portal_type                = 'QuizQA'    
    global_allow               = False
    filter_content_types       = False
    allowed_content_types      = ()
    allow_discussion           = False
    content_icon               = 'quizqa.gif'
    immediate_view             = 'quizqa_view'
    default_view               = 'quizqa_view'
    schema = QuizQASchema

    security.declareProtected(CMFCorePermissions.View, 'getPosition')
    def getPosition(self):
        """Return question/answers position in the quiz
        """
        
        return self.getQuiz().getQuizQAPosition(self.UID())
        
    security.declareProtected(CMFCorePermissions.View, 'tag')
    def tag(self, **kwargs):
        """Generate image tag using the api of the ImageField
        """
        if 'title' not in kwargs:
            kwargs['title'] = self.Title()
        return self.getField('image').tag(self, **kwargs)

    security.declareProtected(CMFCorePermissions.View, 'getCorrectAnswerIds')
    def getCorrectAnswerIds(self):
        """Returns ids of correct answers
        First question had id 0. Next has id 1 and go on.
        """
        
        answers = self.getAnswers()
        return [x for x in range(0, len(answers)) if answers[x][1]]
        
    security.declareProtected(CMFCorePermissions.View, 'getCorrectAnswersCount')
    def getCorrectAnswersCount(self):
        """Returns the number of correct answers defined"""
        
        return len(self.getCorrectAnswerIds())

    security.declareProtected(CMFCorePermissions.View, 'getAnswerTitles')
    def getAnswerTitles(self, answer_ids):
        """Returns answer titles
        
        @param answer_ids: Ids of answers you want to get Title
        """
        
        answers = self.getAnswers()
        return [answers[x][0] for x in range(0, len(answers)) if x in answer_ids]

    security.declareProtected(CMFCorePermissions.View, 'getResult')
    def getResult(self):
        """Returns a dictionnary
        - uid : uid of QuizQA object
        - answer_ids: ids of user answers
        - valid: Are answers valid ?Returns true if list of answers is valid
        """
        
        # Get answers from request
        quiz_qa_uid = self.UID()
        request = self.REQUEST
        key = 'quiz_qa_%s' % quiz_qa_uid
        answer_ids = request.get(key, ())
        
        # Build structure
        item = {}
        item['uid'] = quiz_qa_uid
        item['answer_ids'] = answer_ids
        correct_answer_ids =  self.getCorrectAnswerIds()
        item['valid'] = False
        if len(answer_ids) == len(correct_answer_ids):
            item['valid'] = (len([x for x in correct_answer_ids if x in answer_ids]) == len(correct_answer_ids))
        return item

    security.declareProtected(CMFCorePermissions.View, 'getAnswerMessage')
    def getAnswerMessage(self, is_valid):
        """Returns answer message. Use quiz one if not defined on QuizQA
        
        @param is_valid: Returns right message if this parameter is True 
            otherwise returns wrong message
        """
        
        if is_valid:
            # Returns the message for right answer
            msg = self.getRightAnswerMessage()
        else:
            # Returns the message for wrong answer
            msg = self.getWrongAnswerMessage()
        
        if not msg:
            return self.getQuiz().getAnswerMessage(is_valid)
        
        return msg


    ##code-section class-header #fill in your manual code here
    ##/code-section class-header
    security.declareProtected(CMFCorePermissions.View, 'post_validate')
    def post_validate(self, REQUEST=None, errors=None):
        """Extra validation
        """
        
        # Check answers
        answers = REQUEST.form.get('answers', None)
        error = False
        if answers is None:
            error = True
        elif len(answers) < 2:
            error = True
        elif len([x for x in answers if x.get('checked', False)]) < 1:
            error = True
            
        if error:
            errors['answers'] = 'message_quiz_answers_validator'
        return

registerType(QuizQA)



